# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Shared functions for the buildbot master and slave"""

__metaclass__ = type
__all__ = [
    'buildbot_reconfig',
    'buildslave_start',
    'buildslave_stop',
    'config_json',
    'create_slave',
    'generate_string',
    'HTTP_PORT_PROTOCOL',
    'slave_json',
    ]

import os
import subprocess
import uuid

from helpers import (
    get_config,
    log,
    run,
    Serializer,
    )


HTTP_PORT_PROTOCOL = '8010/TCP'


def _get_buildbot_dir():
    config = get_config()
    return config.get('installdir')


def generate_string(prefix=""):
    """Generate a unique string and return it."""
    return prefix + uuid.uuid4().hex


def buildbot_create(buildbot_dir):
    """Create a buildbot instance in `buildbot_dir`."""
    if not os.path.exists(os.path.join(buildbot_dir, 'buildbot.tac')):
        return run('buildbot', 'create-master', buildbot_dir)


def buildbot_running(buildbot_dir):
    pidfile = os.path.join(buildbot_dir, 'twistd.pid')
    if os.path.exists(pidfile):
        buildbot_pid = open(pidfile).read().strip()
        try:
            # Is buildbot running?
            run('kill', '-0', buildbot_pid)
        except subprocess.CalledProcessError:
            return False
        return True
    return False


def buildbot_stop():
    buildbot_dir = _get_buildbot_dir()
    if buildbot_running(buildbot_dir):
        # Buildbot is running, stop it.
        log('Stopping buildbot')
        run('buildbot', 'stop', buildbot_dir)


def buildbot_reconfig():
    buildbot_dir = _get_buildbot_dir()
    pidfile = os.path.join(buildbot_dir, 'twistd.pid')
    running = False
    if os.path.exists(pidfile):
        buildbot_pid = open(pidfile).read().strip()
        try:
            # Is buildbot running?
            run('kill', '-0', buildbot_pid)
        except subprocess.CalledProcessError:
            # Buildbot isn't running, so no need to reconfigure it.
            pass
        else:
            # Buildbot is running, reconfigure it.
            log('--> Reconfiguring buildbot')
            # XXX as root! :-(
            # reconfig is broken in 0.8.3 (Oneiric)
            # run('buildbot', 'reconfig', buildbot_dir)
            run('buildbot', 'stop', buildbot_dir)
            run('buildbot', 'start', buildbot_dir)
            log('<-- Reconfiguring buildbot')
            running = True
    if not running:
        # Buildbot isn't running so start afresh.
        if os.path.exists(os.path.join(buildbot_dir, 'master.cfg')):
            log('--> Starting buildbot')
            # XXX as root! :-(
            run('buildbot', 'start', buildbot_dir)
            log('<-- Starting buildbot')


def _get_tac_filename(buildbot_dir):
    return os.path.join(buildbot_dir, 'buildbot.tac')


def buildslave_stop(buildbot_dir=None):
    if buildbot_dir is None:
        buildbot_dir = _get_buildbot_dir()
    exit_code = subprocess.call(['buildslave', 'stop', buildbot_dir])
    tac_file = _get_tac_filename(buildbot_dir)
    if os.path.exists(tac_file):
        os.remove(tac_file)
    return exit_code


def buildslave_start(buildbot_dir=None):
    if buildbot_dir is None:
        buildbot_dir = _get_buildbot_dir()
    return subprocess.call(['buildslave', 'start', buildbot_dir])


def create_slave(name, passwd, host='localhost', buildbot_dir=None):
    if buildbot_dir is None:
        buildbot_dir = _get_buildbot_dir()
    if not os.path.exists(buildbot_dir):
        os.makedirs(buildbot_dir)
    return subprocess.call([
        'buildslave', 'create-slave', buildbot_dir, host, name, passwd])


slave_json = Serializer('/tmp/slave_info.json')
config_json = Serializer('/tmp/config.json')
