# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Helper functions for writing hooks in python."""

__metaclass__ = type
__all__ = [
    'apt_get_install',
    'command',
    'DictDiffer',
    'get_config',
    'get_value_from_line',
    'grep',
    'install_extra_repository',
    'log',
    'log_entry',
    'log_exit',
    'run',
    'relation_get',
    'relation_set',
    'unit_info',
    ]

from textwrap import dedent
import base64
import json
import os
import re
import subprocess
import sys
import time
import yaml


def run(*args):
    """Run the command with the given arguments.

    The first argument is the path to the command to run, subsequent arguments
    are command-line arguments to be passed.
    """
    process = subprocess.Popen(args, stdout=subprocess.PIPE,
        stderr=subprocess.PIPE, close_fds=True)
    stdout, stderr = process.communicate()
    if process.returncode:
        raise subprocess.CalledProcessError(
            process.returncode, repr(args), output=stdout+stderr)
    return stdout


def command(*base_args):
    """Return a callable that will run the given command with any arguments.

    The first argument is the path to the command to run, subsequent arguments
    are command-line arguments to "bake into" the returned callable.

    The callable runs the given executable and also takes arguments that will
    be appeneded to the "baked in" arguments.

    For example, this code will list a file named "foo" (if it exists):

        ls_foo = command('/bin/ls', 'foo')
        ls_foo()

    While this invocation will list "foo" and "bar" (assuming they exist):

        ls_foo('bar')
    """
    def callable_command(*args):
        all_args = base_args + args
        return run(*all_args)

    return callable_command


log = command('juju-log')
apt_get_install = command('apt-get', 'install', '-y', '--force-yes')


def get_config():
    config_get = command('config-get', '--format=json')
    return json.loads(config_get())


def install_extra_repository(extra_repository):
    try:
        run('apt-add-repository', extra_repository)
        run('apt-get', 'update')
    except subprocess.CalledProcessError as e:
        log('Error adding repository: ' + extra_repository)
        log(e)
        raise


def relation_get(*args):
    cmd = command('relation-get')
    return cmd(*args).strip()


def relation_set(**kwargs):
    cmd = command('relation-set')
    args = ['{}={}'.format(k, v) for k, v in kwargs.items()]
    return cmd(*args)


def encode_file(filename):
    """base64 encode the contents of a file and return it."""
    with open(filename) as f:
        contents = f.read()
    return base64.b64encode(contents)


def grep(content, filename):
    with open(filename) as f:
        for line in f:
            if re.match(content, line):
                return line.strip()


def get_value_from_line(line):
    return line.split('=')[1].strip('"\' ')


def script_name():
    return os.path.basename(sys.argv[0])


def log_entry():
    log("<-- Entering {}".format(script_name()))


def log_exit():
    log("--> Exiting {}".format(script_name()))


class DictDiffer:
    """
    Calculate the difference between two dictionaries as:
    (1) items added
    (2) items removed
    (3) keys same in both but changed values
    (4) keys same in both and unchanged values
    """

    # Based on answer by hughdbrown at:
    # http://stackoverflow.com/questions/1165352

    def __init__(self, current_dict, past_dict):
        self.current_dict = current_dict
        self.past_dict = past_dict
        self.set_current = set(current_dict)
        self.set_past = set(past_dict)
        self.intersect = self.set_current.intersection(self.set_past)

    @property
    def added(self):
        return self.set_current - self.intersect

    @property
    def removed(self):
        return self.set_past - self.intersect

    @property
    def changed(self):
        return set(key for key in self.intersect
                   if self.past_dict[key] != self.current_dict[key])
    @property
    def unchanged(self):
        return set(key for key in self.intersect
                   if self.past_dict[key] == self.current_dict[key])
    @property
    def modified(self):
        return self.current_dict != self.past_dict

    @property
    def added_or_changed(self):
        return self.added.union(self.changed)

    def _changes(self, keys):
        new = {}
        old = {}
        for k in keys:
            new[k] = self.current_dict.get(k)
            old[k] = self.past_dict.get(k)
        return "%s -> %s" % (old, new)

    def __str__(self):
        if self.modified:
            s = dedent("""\
            added: %s
            removed: %s
            changed: %s
            unchanged: %s""") % (
                self._changes(self.added),
                self._changes(self.removed),
                self._changes(self.changed),
                list(self.unchanged))
        else:
            s = "no changes"
        return s


def unit_info(service_name, item_name, data=None):
    if data is None:
        data = yaml.safe_load(run('juju', 'status'))
    service = data['services'].get(service_name)
    if service is None:
        # XXX 2012-02-08 gmb:
        #     This allows us to cope with the race condition that we
        #     have between deploying a service and having it come up in
        #     `juju status`. We could probably do with cleaning it up so
        #     that it fails a bit more noisily after a while.
        return ''
    units = service['units']
    item = units.items()[0][1][item_name]
    return item


def wait_for_unit(service_name, timeout=120):
    start_time = time.time()
    while True:
        state = unit_info(service_name, 'state')
        if 'error' in state or state == 'started':
            break
        if time.time() - start_time >= timeout:
            raise RuntimeError('timeout waiting for service to start')
        time.sleep(0.1)
    if state != 'started':
        raise RuntimeError('unit did not start, state: ' + state)


def wait_for_relation(service_name, relation_name, timeout=120):
    start_time = time.time()
    while True:
        relations = unit_info(service_name, 'relations')
        if relations[relation_name]['state'] == 'up':
            break
        if time.time() - start_time >= timeout:
            raise RuntimeError('timeout waiting for relation to be up')
        time.sleep(0.1)


class Serializer:
    """Handle JSON (de)serialization."""

    def __init__(self, path, default=None, serialize=None, deserialize=None):
        self.path = path
        self.default = default or {}
        self.serialize = serialize or json.dump
        self.deserialize = deserialize or json.load

    def exists(self):
        return os.path.exists(self.path)

    def get(self):
        if self.exists():
            with open(self.path) as f:
                return self.deserialize(f)
        return self.default

    def set(self, data):
        with open(self.path, 'w') as f:
            self.serialize(data, f)
